@@ -1,5 +1,10 @@
 from __future__ import annotations
 
+import pickle
+import pprint
+import os
+import copy
+import json
 from numpy.random import default_rng
 from spotPython.design.spacefilling import spacefilling
 from spotPython.build.kriging import Kriging
@@ -24,6 +29,7 @@ from spotPython.utils.init import fun_control_init, optimizer_control_init, surr
 from spotPython.utils.compare import selectNew
 from spotPython.utils.aggregate import aggregate_mean_var, select_distant_points
 from spotPython.utils.repair import remove_nan
+from spotPython.utils.file import get_experiment_filename
 from spotPython.budget.ocba import get_ocba_X
 import logging
 import time
@@ -35,6 +41,7 @@ from spotPython.hyperparameters.values import (
 )
 import plotly.graph_objects as go
 from typing import Callable
+from spotPython.utils.numpy2json import NumpyEncoder
 
 
 logger = logging.getLogger(__name__)
@@ -248,6 +255,7 @@ class Spot:
         self.infill_criterion = self.fun_control["infill_criterion"]
         self.n_points = self.fun_control["n_points"]
         self.max_surrogate_points = self.fun_control["max_surrogate_points"]
+        self.progress_file = self.fun_control["progress_file"]
 
         # if the key "spot_writer" is not in the dictionary fun_control,
         # set self.spot_writer to None else to the value of the key "spot_writer"
@@ -612,6 +620,114 @@ class Spot:
             logger.warning("No new XO found on surrogate. Generate new solution %s", X0)
             return X0
 
+    def de_serialize_dicts(self) -> tuple:
+        """
+        Deserialize the spot object and return the dictionaries.
+
+        Args:
+            self (object):
+                Spot object
+
+        Returns:
+            (tuple):
+                tuple containing dictionaries of spot object:
+                fun_control (dict): function control dictionary,
+                design_control (dict): design control dictionary,
+                optimizer_control (dict): optimizer control dictionary,
+                spot_tuner_control (dict): spot tuner control dictionary, and
+                surrogate_control (dict): surrogate control dictionary
+        """
+        spot_tuner = copy.deepcopy(self)
+        spot_tuner_control = vars(spot_tuner)
+
+        fun_control = copy.deepcopy(spot_tuner_control["fun_control"])
+        design_control = copy.deepcopy(spot_tuner_control["design_control"])
+        optimizer_control = copy.deepcopy(spot_tuner_control["optimizer_control"])
+        surrogate_control = copy.deepcopy(spot_tuner_control["surrogate_control"])
+
+        # remove keys from the dictionaries:
+        spot_tuner_control.pop("fun_control", None)
+        spot_tuner_control.pop("design_control", None)
+        spot_tuner_control.pop("optimizer_control", None)
+        spot_tuner_control.pop("surrogate_control", None)
+        spot_tuner_control.pop("spot_writer", None)
+        spot_tuner_control.pop("design", None)
+        spot_tuner_control.pop("fun", None)
+        spot_tuner_control.pop("optimizer", None)
+        spot_tuner_control.pop("rng", None)
+        spot_tuner_control.pop("surrogate", None)
+
+        fun_control.pop("core_model", None)
+        fun_control.pop("metric_river", None)
+        fun_control.pop("metric_sklearn", None)
+        fun_control.pop("metric_torch", None)
+        fun_control.pop("prep_model", None)
+        fun_control.pop("spot_writer", None)
+        fun_control.pop("test", None)
+        fun_control.pop("train", None)
+
+        surrogate_control.pop("model_optimizer", None)
+        surrogate_control.pop("surrogate", None)
+
+        return (fun_control, design_control, optimizer_control, spot_tuner_control, surrogate_control)
+
+    def write_db_dict(self) -> None:
+        """Writes a dictionary with the experiment parameters to the json file spotPython_db.json.
+
+        Args:
+            self (object): Spot object
+
+        Returns:
+            (NoneType): None
+
+        """
+        # get the time in seconds from 1.1.1970 and convert the time to a string
+        t_str = str(time.time())
+        ident = str(self.fun_control["PREFIX"]) + "_" + t_str
+
+        (
+            fun_control,
+            design_control,
+            optimizer_control,
+            spot_tuner_control,
+            surrogate_control,
+        ) = self.de_serialize_dicts()
+        print("\n**********************")
+        print("The following dictionaries are written to the json file spotPython_db.json:")
+        print("fun_control:")
+        pprint.pprint(fun_control)
+        print("design_control:")
+        pprint.pprint(design_control)
+        print("optimizer_control:")
+        pprint.pprint(optimizer_control)
+        print("spot_tuner_control:")
+        pprint.pprint(spot_tuner_control)
+        print("surrogate_control:")
+        pprint.pprint(surrogate_control)
+        db_dict = {
+            str(ident): {
+                "fun_control": fun_control,
+                "design_control": design_control,
+                "surrogate_control": surrogate_control,
+                "optimizer_control": optimizer_control,
+                "spot_tuner_control": spot_tuner_control,
+            }
+        }
+
+        # check if the directory "db_dicts" exists.
+        if not os.path.exists("db_dicts"):
+            try:
+                os.makedirs("db_dicts")
+            except OSError as e:
+                raise Exception(f"Error creating directory: {e}")
+        if os.path.exists("db_dicts"):
+            try:
+                with open("db_dicts/" + self.fun_control["db_dict_name"], "a") as f:
+                    json.dump(db_dict, f, indent=4, cls=NumpyEncoder)
+                f.close()
+            except OSError as e:
+                raise Exception(f"Error writing to file: {e}")
+
     def run(self, X_start=None) -> Spot:
         self.initialize_design(X_start)
         # New: self.update_stats() moved here:
@@ -640,6 +756,10 @@ class Spot:
         if self.spot_writer is not None:
             writer = self.spot_writer
             writer.close()
+        pprint.pprint(self.fun_control)
+        if self.fun_control["db_dict_name"] is not None:
+            self.write_db_dict()
+        self.save_experiment()
         return self
 
     def initialize_design(self, X_start=None) -> None:
@@ -928,12 +1048,25 @@ class Spot:
             self.plot_model()
 
     def show_progress_if_needed(self, timeout_start) -> None:
+        """Show progress bar if `show_progress` is `True`. If
+        self.progress_file is not `None`, the progress bar is saved
+        in the file with the name `self.progress_file`.
+
+        Args:
+            self (object): Spot object
+            timeout_start (float): start time
+
+        Returns:
+            (NoneType): None
+        """
         if not self.show_progress:
             return
         if isfinite(self.fun_evals):
-            progress_bar(progress=self.counter / self.fun_evals, y=self.min_y)
+            progress_bar(progress=self.counter / self.fun_evals, y=self.min_y, filename=self.progress_file)
         else:
-            progress_bar(progress=(time.time() - timeout_start) / (self.max_time * 60), y=self.min_y)
+            progress_bar(
+                progress=(time.time() - timeout_start) / (self.max_time * 60), y=self.min_y, filename=self.progress_file
+            )
 
     def generate_design(self, size, repeats, lower, upper) -> np.array:
         return self.design.scipy_lhd(n=size, repeats=repeats, lower=lower, upper=upper)
@@ -1863,3 +1996,34 @@ class Spot:
         if show:
             fig.show()
         return fig
+
+    def save_experiment(self, filename=None) -> None:
+        """
+        Save the experiment to a file.
+        """
+        design_control = copy.deepcopy(self.design_control)
+        fun_control = copy.deepcopy(self.fun_control)
+        optimizer_control = copy.deepcopy(self.optimizer_control)
+        spot_tuner = copy.deepcopy(self)
+        surrogate_control = copy.deepcopy(self.surrogate_control)
+
+        # remove the key "spot_writer" from the fun_control dictionary,
+        # because it is not serializable.
+        # TODO: It will be re-added when the experiment is loaded.
+        fun_control.pop("spot_writer", None)
+        experiment = {
+            "design_control": design_control,
+            "fun_control": fun_control,
+            "optimizer_control": optimizer_control,
+            "spot_tuner": spot_tuner,
+            "surrogate_control": surrogate_control,
+        }
+        # check if the key "spot_writer" is in the fun_control dictionary
+        if "spot_writer" in fun_control and fun_control["spot_writer"] is not None:
+            fun_control["spot_writer"].close()
+        PREFIX = fun_control["PREFIX"]
+        if filename is None and PREFIX is not None:
+            filename = get_experiment_filename(PREFIX)
+        if filename is not None:
+            with open(filename, "wb") as handle:
+                pickle.dump(experiment, handle, protocol=pickle.HIGHEST_PROTOCOL)
